"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const ejs = require("ejs");
const html_minifier_1 = require("html-minifier");
const utils_1 = require("./utils");
function getAssetName(rawName) {
    let name = rawName;
    if (name.includes("/")) {
        let list = name.split("/");
        name = list[list.length - 1];
    }
    return name.split(".")[0];
}
function readFileContent(fileName) {
    return fs.readFileSync(fileName).toString();
}
function insertPreload({ rel, asValue, includes }, chunks, htmlContent, entryName) {
    let preloadChunks = [];
    chunks = chunks.filter(chunk => !chunk.files[0].endsWith(".map"));
    if (includes === "all" || (typeof includes === "object" && includes[entryName])) {
        const entryChunk = chunks.find(chunk => chunk.initial && chunk.names[0] === entryName);
        if (typeof entryChunk !== "undefined") {
            chunks = chunks.filter(chunk => chunk.parents.includes(entryChunk.id));
        }
        else {
            chunks = [];
        }
        if (includes !== "all") {
            includes = includes[entryName];
        }
    }
    if (includes === "all") {
        preloadChunks = chunks.map(chunk => ({
            name: chunk.names[0],
            file: chunk.files[0]
        }));
    }
    else if (Array.isArray(includes)) {
        preloadChunks = chunks
            .map(chunk => ({
            name: chunk.names[0],
            file: chunk.files[0]
        }))
            .filter(chunk => includes.includes(chunk.name));
    }
    if (preloadChunks.length === 0) {
        return htmlContent;
    }
    const appendHtml = preloadChunks
        .map(({ file, name }) => {
        if (rel === "preload") {
            if (typeof asValue === "undefined") {
                if (file.endsWith(".css")) {
                    asValue = "style";
                }
                else if (file.endsWith("woff2")) {
                    asValue = "font";
                }
                else {
                    asValue = "script";
                }
            }
            const crossOrigin = asValue === "font" ? `crossorigin="crossorigin" ` : "";
            return `<link rel="preload" as="${asValue}" href=${file} ${crossOrigin}>`;
        }
        return `<link rel="${rel}" href=${file} >`;
    })
        .join("\n");
    return htmlContent.replace("</head>", appendHtml + "</head>");
}
class HtmlPlugin {
    constructor(options) {
        this.options = Object.assign({
            template: path.resolve(__dirname, "../config/index.html"),
            title: "ezbuy"
        }, options);
    }
    apply(compiler) {
        compiler.plugin("emit", (compilation, cb) => {
            const { entry, template, entryTemplates, preload, disableMinify } = this.options;
            const { assets, chunks } = compilation.getStats().toJson();
            const allAssets = chunks
                .filter(chunk => chunk.initial)
                .map(chunk => chunk.files[0]);
            const entryAssets = [];
            let shareAssets = [];
            allAssets.forEach(asset => {
                const name = getAssetName(asset);
                if (entry[name]) {
                    entryAssets.push({
                        name,
                        content: asset
                    });
                }
                else {
                    // common entry assets
                    shareAssets.push(asset);
                }
            });
            const manifestIndex = shareAssets.findIndex(item => item.includes("manifest"));
            let manifestContent = "";
            if (manifestIndex !== -1) {
                const manifestItem = shareAssets.splice(manifestIndex, 1)[0];
                manifestContent = compilation.assets[manifestItem].source();
                delete compilation.assets[manifestItem];
            }
            // get dll entry
            const dllAssets = assets
                .filter(asset => asset.name.includes(".dll.js"))
                .map(asset => asset.name)
                .sort((a) => (a.startsWith("common") ? -1 : 1));
            const shareScripts = shareAssets.concat(dllAssets);
            const shareStyles = assets
                .filter(asset => asset.name.includes(".css"))
                .map(asset => asset.name);
            let defaultTemplateContent = "";
            if (template) {
                defaultTemplateContent = readFileContent(template);
            }
            else {
                // 正常的逻辑不会进入这里
                utils_1.printError("The default template is empty.");
                return;
            }
            const scripts = entryAssets.map(item => {
                const templateContent = typeof entryTemplates !== "undefined" && typeof entryTemplates[item.name] !== "undefined"
                    ? readFileContent(entryTemplates[item.name])
                    : defaultTemplateContent;
                let htmlContent = ejs.render(templateContent, {
                    title: this.options.title,
                    styles: shareStyles,
                    scripts: shareScripts.concat(item.content),
                    manifestContent,
                    customTplData: this.options.customTplData
                });
                if (this.options.isProd && typeof preload !== "undefined") {
                    htmlContent = insertPreload(preload, chunks, htmlContent, item.name);
                }
                return {
                    name: item.name,
                    content: disableMinify !== true && this.options.isProd
                        ? html_minifier_1.minify(htmlContent, {
                            removeAttributeQuotes: true,
                            removeComments: true,
                            collapseWhitespace: true
                        })
                        : htmlContent
                };
            });
            scripts.forEach(item => {
                compilation.assets[`${item.name}.html`] = {
                    source: () => item.content,
                    size: () => item.content.length
                };
            });
            return cb();
        });
    }
}
exports.default = HtmlPlugin;
//# sourceMappingURL=HtmlPlugin.js.map