"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const webpack = require("webpack");
const chalk = require("chalk");
const CacheData_1 = require("./CacheData");
const events_1 = require("events");
const webpackDevMiddleware = require("webpack-dev-middleware");
const express = require("express");
const ezmok = require("ezmok");
const webpack_1 = require("./utils/webpack");
const utils_1 = require("./utils");
const host = "0.0.0.0";
const ADDED = Symbol("added");
const BUILDING = Symbol("building");
const BUILT = Symbol("built");
class Invalidator {
    constructor(devMiddleware) {
        this.devMiddleware = devMiddleware;
        this.building = false;
        this.rebuildAgain = false;
    }
    invalidate() {
        if (this.building) {
            this.rebuildAgain = true;
            return;
        }
        this.building = true;
        this.devMiddleware.invalidate();
    }
    startBuilding() {
        this.building = true;
    }
    doneBuilding() {
        this.building = false;
        if (this.rebuildAgain) {
            this.rebuildAgain = false;
            this.invalidate();
        }
    }
}
function getRandomKey(keyList) {
    const randomIndex = Math.floor(Math.random() * keyList.length);
    return keyList[randomIndex];
}
function startServer(webpackConfig, serverConfig) {
    const originEntries = webpackConfig.entry;
    const initEntry = {};
    const builtEntries = {};
    function addToInitEntry(entryKey) {
        if (originEntries[entryKey] && !initEntry[entryKey]) {
            initEntry[entryKey] = originEntries[entryKey];
        }
    }
    if (serverConfig.initEntry) {
        serverConfig.initEntry.forEach(addToInitEntry);
    }
    const cacheEntry = CacheData_1.devEntryCache.getEntry();
    if (cacheEntry.length > 0) {
        cacheEntry.forEach(addToInitEntry);
    }
    // if no entry, generate random one
    if (Object.keys(initEntry).length === 0) {
        const k = getRandomKey(Object.keys(originEntries));
        addToInitEntry(k);
    }
    webpackConfig.entry = initEntry;
    Object.keys(initEntry).forEach(k => {
        builtEntries[k] = {
            status: BUILDING
        };
    });
    const compiler = webpack(webpackConfig);
    const devMiddleware = webpackDevMiddleware(compiler, {
        quiet: true,
        headers: {
            "Access-Control-Allow-Origin": "*"
        }
    });
    const app = express();
    const doneCallbacks = new events_1.EventEmitter();
    const invalidator = new Invalidator(devMiddleware);
    compiler.plugin("make", function addEntry(compilation, done) {
        invalidator.startBuilding();
        const allEntries = Object.keys(builtEntries).map(page => {
            const entry = originEntries[page];
            builtEntries[page].status = BUILDING;
            CacheData_1.devEntryCache.addEntry(page);
            return webpack_1.addWebpackEntry(compilation, this.context, page, entry);
        });
        Promise.all(allEntries)
            .then(() => done())
            .catch(done);
    });
    compiler.plugin("done", function emitWebpackDone() {
        Object.keys(builtEntries).forEach(page => {
            const entryInfo = builtEntries[page];
            if (entryInfo.status !== BUILDING)
                return;
            entryInfo.status = BUILT;
            doneCallbacks.emit(page);
            invalidator.doneBuilding();
        });
    });
    // init express middleware
    app.use(ezmok({
        defaultApiOrigin: serverConfig.defaultProxyOrigin,
        defPath: serverConfig.apiDefPath,
        mockDirectory: CacheData_1.HOME_DIR
    }));
    app.use(function ensurePage(req, res, next) {
        const requestPath = req.path === "/" ? "/index.html" : req.path;
        const url = requestPath.split(".");
        if (url[1] === "html") {
            const entryName = url[0].substr(1);
            if (!originEntries[entryName]) {
                next();
                return;
            }
            const entryInfo = builtEntries[entryName];
            if (entryInfo) {
                if (entryInfo.status === BUILT) {
                    next();
                }
                else if (entryInfo.status === BUILDING) {
                    doneCallbacks.on(entryName, next);
                }
                return;
            }
            utils_1.printWarn(`> Building page: ${entryName}`);
            builtEntries[entryName] = {
                status: ADDED
            };
            doneCallbacks.on(entryName, next);
            invalidator.invalidate();
        }
        else {
            next();
        }
    });
    app.use(devMiddleware);
    app.use(require("webpack-hot-middleware")(compiler, {
        log: () => { }
    }));
    // config static folder
    const { contentBase, staticOptions, port } = serverConfig;
    if (Array.isArray(contentBase)) {
        contentBase.forEach(root => {
            if (typeof root === "string") {
                app.use(express.static(root, staticOptions));
            }
            else {
                app.use(root.prefix, express.static(root.path, staticOptions));
            }
        });
    }
    else if (contentBase) {
        if (typeof contentBase === "string") {
            app.use(express.static(contentBase, staticOptions));
        }
        else {
            app.use(contentBase.prefix, express.static(contentBase.path, staticOptions));
        }
    }
    app.locals.env = process.env.NODE_ENV;
    devMiddleware.waitUntilValid(() => {
        const host = `    http://localhost:${port}`;
        console.log(chalk.yellow(`Available entry:`));
        console.log(chalk.yellow(`${host}/ezmok`));
        Object.keys(webpackConfig.entry).forEach(k => {
            console.log(chalk.yellow(`${host}/${k}.html`));
        });
        console.log();
    });
    console.log("> Starting dev server...");
    app.listen(port, host);
}
exports.default = startServer;
//# sourceMappingURL=startServer.js.map